/*
 * distroquery_api - API for querying data generated by distromatic
 *
 * Copyright (C) 2024 by Silvan Calarco <silvan.calarco@mambasoft.it>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include "DistroqueryAPI.hpp"
#include <functions.h>

DistroqueryAPI::DistroqueryAPI(configTag* conf): config(conf) {
}

json DistroqueryAPI::configToJsonRepositories() {
    json j;
    struct configTag *ct = config;
    while (ct) {
        json archs = {};
        for (auto arch: ct->arch) {
            if (arch == NULL) break;
            archs.push_back(arch);
        }
        json parents = json::array();
        for (auto parent: ct->repository) {
            if (parent == NULL) break;
            if (parent->tag != ct->tag) parents.push_back(parent->tag);
        }
        json repository = {
            { "tag", ct->tag },
            { "archs", archs },
            { "description", ct->description },
            { "parents", parents }
        };
        j.push_back(repository);
        ct = ct->next;
    }
    return j;
}

void DistroqueryAPI::sendErrorResponse(string message) {
    json j;
    j["error"] = message;
    cout << j.dump();
    exit(0);
}

json DistroqueryAPI::getPackageSourceDetailsById(configTag* ct, long id) {
    json j;
    string sql;
    sqlite3_stmt *stmt;

    auto db = openRepositoryDatabase(ct);
    if (!db) {
        j["error"] = "error opening sources database for repository " + string(ct->tag);
        return j;
    }
    sql = "SELECT * FROM sources,packagers WHERE sources.id=" + to_string(id) +
        " AND sources.id_packager=packagers.id";
    if (!sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt, NULL) == SQLITE_OK) {
        j["error"] = "error preparing query '" + sql + "'";
        return j;
    }
    if (sqlite3_step(stmt) == SQLITE_ROW) {
        j["id"] = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "id"));
        j["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "name")));
        j["summary"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "summary")));
        j["epoch"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "epoch")));
        j["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "version")));
        j["release"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "release")));
        j["description"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "description")));
        j["group"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "groupdescr")));
        j["license"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "license")));
        j["url"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "url")));
        //get_favicon_from_url((const char*)sqlite3_column_text(stmt1,sqlite3_find_column_id(stmt1, NULL, "url")),buffer,PATH_MAX);
        j["size"] = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "size"));
        j["maintainer"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, "packagers", "name")));
        // Convert buildtime to ISO-8601
        auto itt = (time_t)sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "buildtime"));
        ostringstream ss;
        ss << std::put_time(gmtime(&itt), "%FT%TZ");
        j["buildtime"] = ss.str();
        // Download URL
        j["download_url"] = string(ct->download_prefix) + ct->download_dir + "/SRPMS.base/" +
            string(j["name"]) + "-" + string(j["version"]) + "-" + string(j["release"]) + ".src.rpm";
        // Source URL
        string name = string(j["name"]);
        j["source_url"] = "https://src.openmamba.org/rpms/" + replaceAll(name, "+", "Plus");
    } else {
        j["error"] = "no results from query '" + sql + "'";
        return j;
    }
    return j;
}

json DistroqueryAPI::getRepositoryPackages(string repository) {
    json j;
    string sql;
    sqlite3_stmt *stmt;

    struct configTag* ct = findRepositoryByTag(repository.c_str());
    if (ct == NULL) {
        j["error"] = "repository with tag '" + repository + "' does not exist";
        return j;
    }

    auto db = openRepositoryDatabase(ct);
    if (!db) {
        j["error"] = "error opening database for repository " + repository;
        return j;
    }

    sql = "SELECT * FROM sources ORDER BY name COLLATE NOCASE ASC";
    if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt, NULL) == SQLITE_OK) {
        j = json::array();
        while (sqlite3_step(stmt) == SQLITE_ROW) {
            json package;
            package["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "name")));
            package["summary"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "summary")));
            j.push_back(package);
        }
        sqlite3_finalize(stmt);
    } else {
        j["error"] = "error preparing query '" + sql + "'";
        return j;
    }
    return j;
}

json DistroqueryAPI::getBuiltPackagesFromSourceID(configTag* ct, long id) {
    json j = json::array();
    string sql;
    sqlite3_stmt *stmt;

    for (auto arch: ct->arch) {
        if (arch == NULL) break;
        auto db = openRepositoryDatabase(ct, arch);
        if (!db) {
            j = {};
            j["error"] = "error opening database for repository '" + string(ct->tag) + "' and arch '" + arch + "'";
            return j;
        }

        sql = "SELECT * FROM packages WHERE id_source = " + to_string(id) + " ORDER BY name";
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt) == SQLITE_ROW) {
                json package = {};
                package["arch"] = arch;
                package["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "name")));
                j.push_back(package);
            }
            sqlite3_finalize(stmt);
        } else {
            j = {};
            j["error"] = "error preparing query '" + sql + "'";
            return j;
        }
    }
    return j;
}

json DistroqueryAPI::getPackageSourceDetails(string repository, string package) {
    json j;
    string sql;
    sqlite3_stmt *stmt, *stmt2;

    struct configTag* ct = findRepositoryByTag(repository.c_str());
    if (ct == NULL) {
        j["error"] = "repository with tag '" + repository + "' does not exist";
        return j;
    }

    auto db = openRepositoryDatabase(ct);
    if (!db) {
        j["error"] = "error opening source database for repository " + repository;
        return j;
    }

    sql = "SELECT * FROM sources WHERE name = '" + package + "'";

    if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt, NULL) == SQLITE_OK) {
        long id;
        if (sqlite3_step(stmt) == SQLITE_ROW) {
            id = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "id"));
            j = getPackageSourceDetailsById(ct, id);
        } else {
            j["error"] = "source package '" + package + "' not found in repository '" + repository + "'";
            return j;
        }

        // Build requires
        sql = "SELECT * FROM buildrequires WHERE id_source=" + to_string(id) +
            " ORDER BY buildrequirename";
        j["buildrequires"] = json::array();
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt2, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt2) == SQLITE_ROW) {
                json buildrequire = {};
                buildrequire["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "buildrequirename")));
                buildrequire["flags"] = rpmSenseFlagsToString(sqlite3_column_int(stmt2,sqlite3_find_column_id(stmt2, NULL, "buildrequireflags")));
                buildrequire["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "buildrequireversion")));
                j["buildrequires"].push_back(buildrequire);
            }
            sqlite3_finalize(stmt2);
        }

        // Children
        j["children"] = getBuiltPackagesFromSourceID(ct, id);

    } else {
        j["error"] = "error preparing query '" + sql + "'";
        return j;
    }
    return j;
}

json DistroqueryAPI::getPackageDetails(string repository, string package, string arch) {
    json j;
    string sql;
    sqlite3_stmt *stmt, *stmt2, *stmt3;

    struct configTag* ct = findRepositoryByTag(repository.c_str());
    if (ct == NULL) {
        j["error"] = "repository with tag '" + repository + "' does not exist";
        return j;
    }

    auto db = openRepositoryDatabase(ct, arch);
    if (!db) {
        j["error"] = "error opening database for repository " + repository + " and arch " + arch;
        return j;
    }
    attachCtDatabases(ct, db, arch);

    sql = "SELECT * FROM packages WHERE name = '" + package + "'";
    if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt, NULL) == SQLITE_OK) {
        long id;
        if (sqlite3_step(stmt) == SQLITE_ROW) {
            id = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "id"));
            j["id"] = id;
            j["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "name")));
            j["summary"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "summary")));
            j["epoch"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "epoch")));
            j["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "version")));
            j["release"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "release")));
            j["description"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "description")));
            j["group"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "groupdescr")));
            string pkg_arch = arch;
            if (sqlite3_find_column_id(stmt, NULL, "arch") > 0) {
                pkg_arch = reinterpret_cast<const char*>(sqlite3_column_text(stmt,sqlite3_find_column_id(stmt, NULL, "arch")));
            }
            j["arch"] = pkg_arch;
            j["size"] = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "size"));

        } else {
            j["error"] = "package '" + package + "' not found in repository '" + repository + "'";
            return j;
        }

        // Source info
        long id_source = sqlite3_column_int(stmt,sqlite3_find_column_id(stmt, NULL, "id_source"));
        j["source"] = getPackageSourceDetailsById(ct, id_source);

        // Download URL
        j["download_url"] = string(ct->download_prefix) + ct->download_dir + "/RPMS." + string(j["arch"]) + "/" +
            string(j["name"]) + "-" + string(j["version"]) + "-" + string(j["release"]) + "." + string(j["arch"]) + ".rpm";

        // Brothers
        sql = "SELECT * FROM packages WHERE id_source=" + to_string(id_source) + " AND NOT id=" + to_string(id);
        j["brothers"] = json::array();
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt2, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt2) == SQLITE_ROW) {
                j["brothers"].push_back(reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "name"))));
            }
            sqlite3_finalize(stmt2);
        }

        // Obsoletes
        sql = "SELECT * FROM obsoletes WHERE id_package=" + to_string(id);
        j["obsoletes"] = json::array();
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt2, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt2) == SQLITE_ROW) {
                json obsolete = {};
                obsolete["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "obsoletename")));
                obsolete["flags"] = rpmSenseFlagsToString(sqlite3_column_int(stmt2,sqlite3_find_column_id(stmt2, NULL, "obsoleteflags")));
                obsolete["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "obsoleteversion")));
                j["obsoletes"].push_back(obsolete);
            }
            sqlite3_finalize(stmt2);
        }

        // Provides
        sql = "SELECT * FROM provides,provided WHERE provides.id_package=" + to_string(id) + 
            " AND provided.id=provides.id_provided ORDER BY provided.name";
        j["provides"] = json::array();
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt2, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt2) == SQLITE_ROW) {
                json provide = {};
                provide["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "name")));
                provide["flags"] = rpmSenseFlagsToString(sqlite3_column_int(stmt2,sqlite3_find_column_id(stmt2, NULL, "provideflags")));
                provide["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "provideversion")));
                j["provides"].push_back(provide);
            }
            sqlite3_finalize(stmt2);
        }

        // Requires
        sql = "SELECT * FROM requires WHERE requires.id_package=" + to_string(id);
//            " AND provided.id=requires.id_provided ORDER BY provided.name";
        j["requires"] = json::array();
        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt2, NULL) == SQLITE_OK) {
            while (sqlite3_step(stmt2) == SQLITE_ROW) {
                json require = {};
                long id_provided = sqlite3_column_int(stmt2,sqlite3_find_column_id(stmt2, NULL, "id_provided"));
                require["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "requirename")));
                require["flags"] = rpmSenseFlagsToString(sqlite3_column_int(stmt2,sqlite3_find_column_id(stmt2, NULL, "requireflags")));
                require["version"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt2,sqlite3_find_column_id(stmt2, NULL, "requireversion")));
                require["providers"] = json::array();
                sql = "SELECT * FROM provides,packages WHERE id_provided=" + to_string(id_provided) +
                    " AND packages.id=provides.id_package";
                if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt3, NULL) == SQLITE_OK) {
                    while (sqlite3_step(stmt3) == SQLITE_ROW) {
                        json provider = {};
                        provider["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt3,sqlite3_find_column_id(stmt3, NULL, "name")));
                        provider["repository"] = ct->tag;
                        require["providers"].push_back(provider);
                    }
                    sqlite3_finalize(stmt3);
                }
                // Resolve requires provided in upstream (alternative) repositories
                int i = 0;
                while (ct->repository[i]) {
                    if (ct->repository[i] != ct) {
                        string altprefix = string(ct->repository[i]->tag) + "_" + arch + ".";
                        sql = "SELECT * FROM " + altprefix + "provides," + altprefix +
                            "packages WHERE providename='" + string(require["name"]) +
                            "' AND " + altprefix + "packages.id=id_package";
                        if (sqlite3_prepare_v2(db, sql.c_str(), sql.length(), &stmt3, NULL) == SQLITE_OK) {
                            while (sqlite3_step(stmt3) == SQLITE_ROW) {
                                json provider = {};
                                provider["name"] = reinterpret_cast<const char*>(sqlite3_column_text(stmt3,sqlite3_find_column_id(stmt3, NULL, "name")));
                                provider["repository"] = ct->repository[i]->tag;
                                require["providers"].push_back(provider);
                            }
                            sqlite3_finalize(stmt3);
                        }
                    }
                    i++;
                }
                j["requires"].push_back(require);
            }
            sqlite3_finalize(stmt2);
        }

    } else {
        j["error"] = "error preparing query '" + sql + "'";
        return j;
    }

    return j;

}

void DistroqueryAPI::getApiResponse(string path_info) {
    cout << "Content-Type: application/json;charset=utf-8\n\n" << endl;

    string path, query_string;
    std::vector<std::string> path_split;

    std::vector<std::string> v = split(path_info, "?");

    if (v.size() > 0) {
        path = v[0];
        path_split = split(path, "/");
    }
    if (v.size() > 1) query_string = v[1];

    if (path_split.size() > 0 ) {
        if (path_split[0] == "repositories") {
            // API service: repositories
            if (path_split.size() != 1)
                sendErrorResponse("expected exactly 1 argument for " + path_split[0]);
            cout << configToJsonRepositories();
        } else if (path_split[0] == "repository") {
            // API service: repository
            if (path_split.size() != 2)
                sendErrorResponse("expected exactly 2 arguments for " + path_split[0]);
            auto packages = getRepositoryPackages(path_split[1]);
            cout << packages.dump();
        } else if (path_split[0] == "package") {
            // API service: package
            if (path_split.size() < 3 || path_split.size() > 4)
                sendErrorResponse("expected 3 or 4 arguments for " + path_split[0]);
            json details;
            string package = urlDecode(path_split[2]);
            if (path_split.size() == 3) {
                details = getPackageSourceDetails(path_split[1], package);
            } else {
                details = getPackageDetails(path_split[1], package, path_split[3]);
            }
            cout << details.dump();
        } else {
            sendErrorResponse("invalid request for service '" + path_split[0]);
        }
    } else {
        sendErrorResponse("empty request");
    }
}
