/*
 * distroquery_functions - Basic functions used by distroquery
 *
 * Copyright (C) 2024 by Silvan Calarco <silvan.calarco@mambasoft.it>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include "distroquery_functions.hpp"
#include "config.h"
#include <string.h>

vector<string> split(string str, string token) {
    vector<string>result;
    while(str.size()){
        auto index = str.find(token);
        if(index!=string::npos){
            result.push_back(str.substr(0,index));
            str = str.substr(index+token.size());
            if(str.size()==0)result.push_back(str);
        }else{
            result.push_back(str);
            str = "";
        }
    }
    return result;
}

string rpmSenseFlagsToString(int flags) {
    string ret;
    if (flags & RPMSENSE_LESS) ret += "<";
    if (flags & RPMSENSE_GREATER) ret += ">";
    if (flags & RPMSENSE_EQUAL) ret += "=";
    return ret;
}

sqlite3* openRepositoryDatabase(struct configTag* ct, string arch, string append) {
    string dbname;
    sqlite3* db;

    if (arch != "" && arch != "sources") {
        dbname = string(ct->repository_dir) + ct->tag + "-" + arch + append + ".db";
        if (sqlite3_open_v2(dbname.c_str(), (sqlite3**)&db, SQLITE_OPEN_READONLY, NULL)) {
            if (db) sqlite3_close(db);
            fprintf(stderr, "ERROR: unable to open sqlite3 db %s\n", dbname.c_str());
            return NULL;
        }
    } else {
        dbname = string(ct->repository_dir) + ct->tag + "-sources" + append + ".db";
        if (sqlite3_open_v2(dbname.c_str(), (sqlite3**)&db, SQLITE_OPEN_READONLY, NULL)) {
            if (db) sqlite3_close(db);
            fprintf(stderr, "ERROR: unable to open sqlite3 db %s\n", dbname.c_str());
            return NULL;
        }
    }
    return db;
}

void attachCtDatabases(struct configTag* ct, sqlite3 *db, string arch) {
    string dbname;
    string sql;
    int i=0;
    char *errmsg;

    while (ct->repository[i]) {
      if (arch != "" && arch != "sources") {
         dbname = string(ct->repository[i]->repository_dir) + ct->repository[i]->tag + "-" + arch + ".db";
         sql = "ATTACH DATABASE '" + dbname + "' as '" + ct->repository[i]->tag + "_" + arch + "'";
         if (sqlite3_exec(db, sql.c_str(), NULL, NULL, &errmsg)) {
            cerr << "ERROR: unable to exec statement for " << sql << ": " << errmsg << endl;
          }
       } else {
         dbname = string(ct->repository[i]->repository_dir) + ct->repository[i]->tag + "-sources.db";
         sql = "ATTACH DATABASE '" + dbname + "' as '" + ct->repository[i]->tag + "_sources'";
         if (sqlite3_exec(db, sql.c_str(), NULL, NULL, &errmsg)) {
            cerr << "ERROR: unable to exec statement for " << sql << ": " << errmsg << endl;
          }
       }
       i++;
    }
}

void attachRepositoryDatabases(struct configTag* ct, sqlite3 *db, string arch, int source) {
   string dbname;
   string sql;
   char *errmsg;

   dbname = string(ct->repository_dir) + ct->tag + "-" + arch + ".db";
   sql = "ATTACH DATABASE '" + dbname + "' as '" + ct->tag + "_" + arch;
   if (sqlite3_exec(db, sql.c_str(), NULL, NULL, &errmsg)) {
      cerr << "ERROR: unable to exec statement for " << sql << ": " << errmsg << endl;
   }
   if (source) {
      dbname = string(ct->repository_dir) + ct->tag + "-sources.db";
      sql = "ATTACH DATABASE '" + dbname + "' as '" + ct->tag + "_sources'";
      if (sqlite3_exec(db, sql.c_str(), NULL, NULL, &errmsg)) {
         cerr << "ERROR: unable to exec statement for " << sql << ":: " << errmsg << endl;
      }
   }
}

int sqlite3_find_column_id(sqlite3_stmt *stmt, const char* table, const char* name) {
   const char* colname;
   int id = 0;

   while ((colname = sqlite3_column_name(stmt, id))) {
      if (!strcmp(colname, name)) {
         if (table) {
            if (!strcmp(sqlite3_column_table_name(stmt, id), table)) return id;
         } else {
            return id;
         }
      }
      id++;
   }
   return -1;
}

string urlDecode(string str) {
    string ret;
    char ch;
    int i, len = str.length();
    unsigned int ii;

    for (i=0; i < len; i++){
        if (str[i] != '%') {
            if (str[i] == '+')
                ret += ' ';
            else
                ret += str[i];
        } else {
            sscanf(str.substr(i + 1, 2).c_str(), "%x", &ii);
            ch = static_cast<char>(ii);
            ret += ch;
            i = i + 2;
        }
    }
    return ret;
}

string replaceAll(string str, const string& from, const string& to) {
    size_t start_pos = 0;
    while((start_pos = str.find(from, start_pos)) != string::npos) {
        str.replace(start_pos, from.length(), to);
        start_pos += to.length(); // Handles case where 'to' is a substring of 'from'
    }
    return str;
}
